package Params;
=head1 NAME

Params.pm - Parameter file parser

$Id: Params.pm,v 1.1 2002/04/12 14:51:34 T.Shigetome Exp $

=head1 SYNOPSIS

	use Params;
	my($rh_parameter);
	my($obj) = new Params("parameter_file_name");
	$rh_parameter = $obj->GetData("type_name");

=head1 DESCRIPTION

Get data from text file.
This is as a common parser of data file.
You can get specified type data.
And you can define any type names.

That data will be got after pre-process.
Pre-processor directives are defined following...

#define, #undef, #if, #elseif, #endif, #include

=head2 METHOD

=over 4

=item PreProcess(I<$filename> [, I<\@buffer>] )

Load from a specified file and keep pre-processed data.
The C<$filename> specify a file name of data.
The C<\@identifier> specify a reference of array that is for keep pre-processed data.
If you omit a reference of array, pre-processed data will be kept in this instance.

	$obj->PreProcess("data.txt", \@buffer);

=item GetData(I<$typename> [, I<\%data>, I<\@identifier>, I<\@buffer>])

Get a reference of hash that is data of specified type from buffer.
The C<$typename> specify type name of your wanted data.
The C<\@identifier> specify reference of array that is for keys of data.
The C<\%data> specify reference of hash that is for data.

	$obj->GetData("type", \%data, \@identifier, \@buffer);

=back

=head2 SYNTAX OF DATA FILE

If first character of each lines is `#' and no match to directive,
 it will be handled as comment.
If end character of each lines is `\', it continue next line.
Each data define is contain one data type, one identifier and one data list.
Data type and identifier is delimited by spaces.
Block of data list is start with `{' and  terminated with "};".
Each data line is terminated with `;' and  must be only one line.

	#comments
	type identifier{
		data-list;
		... ;
		... ;
	};

=over 4

=item I<type>	: data type name

I<nospace character sequence>

=item I<identifier>	:identifier

I<nospace character sequence>

=item I<member-list>	: data list

	{
		nospace-character character-sequence  nospace-character;
		... ;
		... ;
	};

=back

=head2 PRE-PROCESSOR DIRECTIVE

=over 4

=item #define directive

The #define directive is given name to constant variable in data file.

B<#define> I<identifier>
B<#define> I<identifier> I<token string>

Pre-processor will replace `%I<identifier>%' that defined by #define directive.
But, it exclude comment line and replace only already defined value.
Replace process is only onetime this mean can not nest define.
If pre-processor can not find in define, it will search in environment variable.
The identifier can include alphabet, digit number and under score.

=item #undef directive

Undefine identifier that defined with #define directive.

B<#unde>f I<identifier>

=item #if, #elseif, #else, #endif directive

The #if directive is used with #elseif, #else and #endif directives.
It flow control of data parse.
If expression that following #if or #elseif directive is true, following block are valid.
The #if directive have to pair with #endif.
One #if - #endif pair can has any #elseif directive.
The #else directive can exist up to 1 in #if block.
The #if directive can nest.

	#if	expression
		strings
	#elseif	expression
		strings
	#else
		strings
	#endif


In expression you can use the `defined' function.

B<defined>(I<identifier>)

This expression is true, if identifier is defined with #define directive or environment variable.

=item #include directive

The #include directive is load to current line.
Also that loaded file is pre-processed.

B<#include> I<file name>

=back

=head1 AUTHOR

Tetsuya Shigetome, E<lt>t_shigetome@muf.biglobe.ne.jpE<gt>.

=head1 HISTORY

=over 4

=item $Revision: 1.1 $

$Log: Params.pm,v $
Revision 1.1  2002/04/12 14:51:34  T.Shigetome

First release of Perl module version.


=back

=cut

use strict;
use Carp;
no strict 'refs';
require Exporter;
@Params::ISA = qw(Exporter);

# symbols to export by default
@Params::EXPORT = qw( Set2HashRef Set2ArrayRef Set2ScalarRef);
# symbols to export on request
@Params::EXPORT_OK = qw(&Set2HashRef &Set2ArrayRef &Set2ScalarRef $delimiter $start_token $stop_token);

$Params::VERSION = sprintf("%d.%02d", q$Revision: 1.1 $ =~ /(\d+)\.(\d+)/);
$Params::start_token = '\{';
$Params::stop_token = '\};';
$Params::delimiter = ';';
$Params::handle = "ParamsA";
#######################################
# static variable
%Params::DirectiveControl =(
	'if',		\&FlowControlIf,
	'elseif',	\&FlowControlElseIf,
	'else',		\&FlowControlElse,
	'endif',	\&FlowControlEndIf,
	'include',	\&ExecPreProcess,
);
%Params::DefineDirective = (
	'define',	\&Define,
	'undef',	\&UnDefine,
);
######################
# Constructer
sub new
{
my($pkg, $filename) = @_;
my($rh_this);
	$rh_this =  bless {
		'debug'		=>	undef,
		'filename'	=>	$filename,
		'buffer'	=>	undef,
		'ailias'	=>	undef,
		'data'		=>	undef,
	}, $pkg;

	$rh_this->PreProcess() if( defined($filename) );
	return $rh_this;
}
#######################
# Method autoload
sub AUTOLOAD
{
my($pkg) = shift;
#	return if $Whatnew::AUTOLOAD =~ m/::DESTROY$/;
my($type) = ref($pkg) or croak "$pkg is not an object";
my($name) = $Params::AUTOLOAD;
	$name =~ s/.*://;
	croak "Can't access `$name' field in object of class $type" unless( exists($pkg->{$name}) );

	return $pkg->{$name} = shift if( @_ );
	return $pkg->{$name};
}
#	It's a scapegoat for AUTOLOAD.
sub	DESTROY{}
#######################################
# Analyse data that is finishd pre-process
# It's included new-line code.
sub GetData
{
my($pkg, $type, $rh_def_value, $ra_ailias, $ra_buffer) = @_;

	croak "Must specify type" unless( defined($type) );
	unless( defined($ra_buffer) )
	{
		croak "Must need buffered data" unless( defined($pkg->{'buffer'}) );
		$ra_buffer = $pkg->{'buffer'};
	}
	unless( defined($ra_ailias) )
	{
		$pkg->{'ailias'} = [];
		$ra_ailias = $pkg->{'ailias'};
	}
	unless( defined($rh_def_value) )
	{
		$pkg->{'data'} = {};
		$rh_def_value = $pkg->{'data'};
	}
	foreach( grep( /^($type)(\s+)(.*)/,@{$ra_buffer}) )
	{
		if( /(\S+)\s+([^\s$Params::start_token]+)\s*$Params::start_token\s*(.*)\s*$Params::delimiter\s*$Params::stop_token/
		|| /(\S+)\s+([^\s$Params::start_token]+)\s*$Params::start_token\s*(.*)\s*$Params::stop_token/ )
		{
			if( $1 eq $type )
			{# Type check
				# Get ailias
				push(@{$ra_ailias} ,$2);
				# Get data queue
				$rh_def_value->{$ra_ailias->[$#{$ra_ailias}]}=$3;
			}
			else
			{
				warn "Unexcept type $type!=$1\n";
			}
		}
	}
	return	$rh_def_value;
}
#######################################
# Pre-compile that is finished pre-process
# It's included new-line code.
sub PreCompile
{
my($ra_buffer) = @_;
local($_);
my(@pre_buffer);
my($line_buf) = "";
	foreach( @{$ra_buffer} )
	{# Expand parameter define list
		chop($_);
		$line_buf .= $_;
		if( $line_buf =~ /(\S+)\s+([^\s$Params::start_token]+)\s*$Params::start_token\s*(.*)\s*$Params::delimiter\s*$Params::stop_token/
		|| $line_buf =~ /(\S+)\s+([^\s$Params::start_token]+)\s*$Params::start_token\s*(.*)\s*$Params::stop_token/ )
		{
			push(@pre_buffer, $line_buf);
			$line_buf = "";
		}
	}
	return(@pre_buffer);
}
#######################################
# Execute pre-process
# Specified directive are..
# #define
# #undef
# #if,#elseif,#else,#endif
# #include
#
sub PreProcess
{
my($pkg, $filename, $ra_buffer) = @_ ;
local(%Params::defined) = ();
local(@Params::state) = (1);
# normal=0
# search endif=1
# (enter if block or return from if nest)
# search else=2

	$filename = $pkg->{'filename'} unless( defined($filename) );
	unless( defined($ra_buffer) )
	{
		$pkg->{'buffer'} = [];
		$ra_buffer = $pkg->{'buffer'};
	}

	&ExecPreProcess($filename, $ra_buffer, *defined);
	@{$ra_buffer} = &PreCompile($ra_buffer);
	return	wantarray ? @{$ra_buffer} : $ra_buffer;
}
sub ExecPreProcess
{
my($filename) = shift @_;
my($ra_buffer) = shift @_;
local(*defined) =  @_;
my(@temp_buf,@work_buf);
++$Params::handle;

	open($Params::handle,$filename) or die "Can not open $filename: $!";
	# Proceed multiple line
	while( <$Params::handle> )
	{
		if(@temp_buf > 1 && $temp_buf[-1] =~ s/\\\s+$// )
		{
			$temp_buf[-1] .= $_;
		}
		else
		{
			push(@temp_buf,$_);
		}
	}
	close($Params::handle);
	# Delete comment line
	@work_buf = grep( /^\s*([^#]|#\s*[deis])/, @temp_buf);
	# Delete empty line
	@temp_buf = grep( !/^\s*$/, @work_buf);
	foreach( @temp_buf )
	{
		# Delete space character
		s/^\s*(.*)\s*$/$1\n/;
		$_ = &ExtractDefine($_);
		if( /^#\s*(if|else|elseif|endif|include)\s+(.*)/ )
		{# It's a flow control statement
			&{$Params::DirectiveControl{$1}}($2, $ra_buffer, *defined);
			next;
		}
		elsif( $Params::state[$#Params::state] <= 1 )
		{
			if( /^#\s*(define|undef)\s+(.*)/ )
			{
				&{$Params::DefineDirective{$1}}($2);
				next;
			}
			elsif(/^#/)
			{# comment line
				next;
			}
			else
			{
				push(@{$ra_buffer}, $_);
			}
		}
	}
	--$Params::handle;
}
sub ExtractDefine
{
local($_) = @_;

	s/(%\w+%)/&GetDefineString($1)/eg;
#	print $_;
	return	$_;
}
sub GetDefineString
{
local($_) = @_;
	m/%(\w+)%/;
	return($Params::defined{$1}) if( defined($Params::defined{$1}) );
	return($ENV{$1}) if( defined($ENV{$1}));
	return($_);
}
# 0: normal
# 1:Internal True block if EXPR 
# 2:Search Else block(causes if EXPR is False)
# 3:Search #endif (already end True block)
# search endif=1: Internal True block if EXPR 
# and search block end
# (enter if block or return from if nest)
# search else=2: if EXPR is False and search block end
#
# for #else and #elseif
%Params::NextState = (
	0,	"Unexepect #else/elseif\n",
	1,	3,
	2,	1,
	3,	3,
);

sub FlowControlIf
{
my($expr) = @_;
	if( $Params::state[$#Params::state] <= 1 )
	{
		if( &IfDirective($expr) )
		{
			push(@Params::state,1);
		}
		else
		{
			push(@Params::state,2)
		}
	}
	else
	{
		push(@Params::state,3);
	}
}
sub FlowControlElseIf
{
my($expr) = @_;
	if( defined($Params::NextState{$Params::state[$#Params::state]} ) )
	{
		if( $Params::state[$#Params::state] == 2 )
		{
			if( &IfDirective($expr) )
			{
				$Params::state[$#Params::state] = 1;
			}
		}
		else
		{
			$Params::state[$#Params::state] = $Params::NextState{$Params::state[$#Params::state]};
			die $Params::state[$#Params::state] unless ( $Params::state[$#Params::state] =~ /[0-9]+/ );
		}
	}
	else
	{
		die "Unknown state $Params::state[$#Params::state]\n";
	}
}
sub FlowControlElse
{
my($expr) = @_;
	if( defined($Params::NextState{$Params::state[$#Params::state]} ) )
	{
		$Params::state[$#Params::state] = $Params::NextState{$Params::state[$#Params::state]};
		die $Params::state[$#Params::state] unless ( $Params::state[$#Params::state] =~ /[0-9]+/ );
	}
	else
	{
		die "Unknown state $Params::state[$#Params::state]\n";
	}
}
sub FlowControlEndIf
{
my($expr) = @_;
	if( $Params::state[$#Params::state] == 0 )
	{
		die "Unexepect #endif\n";
	}
	pop(@Params::state);
}
#######################################
#	Define/Undefine directive
sub Define
{# Define macro
local($_) = @_;
	m/^(\w+)\s*(.*)/;
	$Params::defined{$1} = $2;
}
sub UnDefine
{
local($_) = @_;
	/^(\w+)/;
	delete $Params::defined{$1};
}
#######################################
#	Flow contorol directive
sub IfDirective
{
local($_) = @_;
local($Params::bool);

	s/defined\(\s*(\w+)\s*\)/(defined(\$Params::defined{"$1"}) || defined(\$ENV{"$1"}))/g;
	eval '$Params::bool = '."$_";warn $@ if $@;
	if( $Params::bool )
	{
		$Params::state = 1;
		return 1;
	}
	else
	{
		$Params::state = 2;
		return 0;
	}
}
#######################################
# Common operation of reference
sub Set2HashRef
{
my($rr_hash,$r_para) = @_;
	if( ref($r_para) eq '' )
	{# No ref
		$r_para = [split(/$Params::delimiter/,$r_para)];
	}
	elsif( ref($r_para) eq 'HASH' )
	{
		$$rr_hash = $r_para;
		return 1;
	}
	elsif( ref($r_para) eq 'SCALAR' )
	{
		$r_para = [$$r_para];
	}
	if( ref($r_para) eq 'ARRAY' )
	{
		foreach( @{$r_para} )
		{
			my($key,$value) = split(/=/,$_);
			$$$rr_hash{$key} = $value;
		}
	}
	else
	{
		return 0;
	}
	return 1;
}
sub Set2ArrayRef
{
my($rr_array,$r_para) = @_;
	if( ref($r_para) eq '' )
	{
		$r_para = [split(/$Params::delimiter/,$r_para)];
	}
	elsif( ref($r_para) eq 'SCALAR' )
	{
		$r_para = [$$r_para];
	}
	if( ref($r_para) eq 'ARRAY' )
	{
		$$rr_array = $r_para;
	}
	else
	{
		return	0;
	}
	return	1;
}
sub Set2ScalarRef
{
my($rr_scalar,$r_para) = @_;
	if( ref($r_para) eq '' )
	{
		$$rr_scalar = $r_para;
	}
	elsif( ref($r_para) eq 'SCALAR' )
	{
		$$rr_scalar = $$r_para;
	}
	else
	{
		return 0;
	}
	return	1;
}
1; #return true 
